package auth

import (
	"context"
	"net/http"
	"net/http/httptest"
	"testing"
)

type authProvider struct{}

func (a authProvider) Valid(ctx context.Context, token string) (*User, error) {
	if token == "this-token-is-valid" {
		return &User{
			Name: "test-user",
		}, nil
	}

	return nil, ErrAuthFailed
}

func TestRequire(t *testing.T) {
	var (
		ap     authProvider
		called bool // Will be set true by appropriate handler
	)

	// No authentication cookie present
	w := httptest.NewRecorder()
	r := httptest.NewRequest("POST", "/", nil)

	hdlr := Require(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// "Success" handler
		t.Error("success handler called, expected authFailed handler")
	}), http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// Auth failed handler
		w.WriteHeader(http.StatusUnauthorized)
		called = true
	}), ap)

	hdlr.ServeHTTP(w, r)

	if http.StatusUnauthorized != w.Code {
		t.Errorf("expected HTTP status %d, got %d", http.StatusUnauthorized, w.Code)
	}

	if !called {
		t.Error("expected handler not called")
	}

	// Invalid cookie in request
	c := http.Cookie{
		Name:  "session",
		Value: "invalid",
	}

	w = httptest.NewRecorder()
	r = httptest.NewRequest("POST", "/", nil)
	r.AddCookie(&c)

	called = false

	hdlr = Require(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// "Success" handler
		t.Error("success handler called, expected authFailed handler")
	}), http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// Auth failed handler
		w.WriteHeader(http.StatusUnauthorized)
		called = true
	}), ap)

	hdlr.ServeHTTP(w, r)

	if http.StatusUnauthorized != w.Code {
		t.Errorf("expected HTTP status %d, got %d", http.StatusUnauthorized, w.Code)
	}

	if !called {
		t.Error("expected handler not called")
	}

	// Valid cookie in request
	c = http.Cookie{
		Name:  "session",
		Value: "this-token-is-valid",
	}

	w = httptest.NewRecorder()
	r = httptest.NewRequest("GET", "/", nil)
	r.AddCookie(&c)

	called = false

	hdlr = Require(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// "Success" handler
		called = true
	}), http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		// Auth failed handler
		w.WriteHeader(http.StatusUnauthorized)
		t.Error("auth failed handler called, expected success handler")
	}), ap)

	hdlr.ServeHTTP(w, r)

	if http.StatusOK != w.Code {
		t.Errorf("expected HTTP status %d, got %d", http.StatusUnauthorized, w.Code)
	}

	if !called {
		t.Error("expected handler not called")
	}
}
