package main

import (
	"bytes"
	"context"
	"database/sql"
	"errors"
	"fmt"
	"html/template"
	"io"
	"net/http"
	"strconv"
	"strings"
	"sync"

	"github.com/go-kit/kit/log/level"

	"git.c3pb.de/gbe/invinoveritas/auth"
	"git.c3pb.de/gbe/invinoveritas/log"
	"git.c3pb.de/gbe/invinoveritas/storage"
	"git.c3pb.de/gbe/invinoveritas/storage/query"
	"git.c3pb.de/gbe/invinoveritas/vino"
)

func (h Handler) img() http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		if r.Method != "GET" {
			httpError(w, r, "bad method", errors.New(r.Method), http.StatusMethodNotAllowed)
			return
		}

		parts := strings.Split(r.URL.Path, "/")

		id, err := strconv.Atoi(parts[len(parts)-1])
		if err != nil {
			httpError(w, r, "can't parse id", err, http.StatusBadRequest)
			return
		}

		img, err := h.Q.LoadPicture(r.Context(), int32(id))
		if err != nil {
			httpError(w, r, "can't load image", err, http.StatusNotFound)
			return
		}

		w.Header().Set("Content-Type", "image/png")

		_, err = io.Copy(w, bytes.NewBuffer(img))
		if err != nil {
			level.Error(log.Get(r)).
				Log("error", err, "msg", "can't write image")

			return
		}
	}
}

func (h Handler) detailsPost(w http.ResponseWriter, r *http.Request) {
	action := r.FormValue("action")
	if strings.HasPrefix(action, "delete-comment-") {
		level.Debug(log.Get(r)).
			Log("action", action, "msg", "handling comment action")

		parts := strings.Split(action, "-")

		commentID, err := strconv.Atoi(parts[len(parts)-1])
		if err != nil {
			httpError(w, r, "bad request", err, http.StatusBadRequest)
			return
		}

		err = h.Q.DeleteComment(r.Context(), int32(commentID))
		if err != nil {
			httpError(w, r, "internal server error", err, http.StatusInternalServerError)
			return
		}

		level.Info(log.Get(r)).
			Log("id", commentID, "msg", "deleted comment")

		http.Redirect(w, r, "/details?id="+r.FormValue("id"), http.StatusSeeOther)
		return
	}

	switch action {
	case "save", "delete":
	default:
		err := fmt.Errorf("action missing or unknown: %q", action)
		httpError(w, r, "invalid parameters", err, http.StatusBadRequest)
		return
	}

	id, err := strconv.Atoi(r.FormValue("id"))
	if err != nil {
		httpError(w, r, "can't parse id", err, http.StatusBadRequest)
		return
	}

	if action == "delete" {
		err = h.Q.DeleteWine(r.Context(), int32(id))
		if err != nil {
			httpError(w, r, "can't delete wine", err, http.StatusInternalServerError)
			return
		}

		http.Redirect(w, r, "/", http.StatusSeeOther)
		return
	}

	name := r.FormValue("name")
	if name == "" {
		httpError(w, r, "name can't be empty", nil, http.StatusBadRequest)
		return
	}

	rawRating := r.FormValue("rating")
	rating, err := strconv.Atoi(rawRating)
	if rawRating != "" && err != nil {
		httpError(w, r, "can't parse rating", nil, http.StatusBadRequest)
		return
	}

	country, err := vino.ISO2CountryCodeFromString(r.FormValue("country"))
	if err != nil {
		httpError(w, r, "can't parse country", nil, http.StatusBadRequest)
		return
	}

	comment := strings.TrimSpace(r.FormValue("comment"))
	if len(comment) > 1024 {
		comment = comment[:1024] + "..."
	}

	err = r.ParseMultipartForm(16 * 1024 * 1024)
	if err != nil {
		httpError(w, r, "can't parse multipart form data", err, http.StatusInternalServerError)
		return
	}

	err = h.Q.RunTx(r.Context(), func(ctx context.Context, q *query.Queries) error {
		if comment != "" {
			err = q.AddComment(ctx, query.AddCommentParams{
				WineID:  int32(id),
				Comment: comment,
			})
			if err != nil {
				httpError(w, r, "can't add comment", err, http.StatusInternalServerError)
				return err
			}
		}

		// See if there's an image file uploaded
		if len(r.MultipartForm.File["picture"]) != 0 {
			picFile, picHdr, err := r.FormFile("picture")
			if err != nil {
				httpError(w, r, "can't open picture file", err, http.StatusInternalServerError)
				return err
			}
			defer picFile.Close()

			err = storage.AddPicture(ctx, q, id, picFile, picHdr.Header.Get("Content-Type"))
			if err != nil {
				httpError(w, r, "can't store picture", err, http.StatusBadRequest)
				return err
			}
		}

		// Update remaining parameters
		params := query.UpdateWineParams{
			WineID: int32(id),
			Name:   name,
			Rating: sql.NullInt32{
				Int32: int32(rating),
				Valid: rawRating != "",
			},
			Country: sql.NullString{
				String: country.String(),
				Valid:  country.Valid(),
			},
		}

		level.Debug(log.Get(r)).Log("update", fmt.Sprint(params))

		err = q.UpdateWine(ctx, params)
		if err != nil {
			httpError(w, r, "can't store update", err, http.StatusInternalServerError)
			return err
		}

		return nil
	})
	if err != nil {
		return
	}

	http.Redirect(w, r, "/details?id="+r.FormValue("id"), http.StatusSeeOther)
}

func (h Handler) details() http.Handler {
	var (
		once sync.Once
		tpl  *template.Template
	)

	type templateData struct {
		Wine     query.LoadWineRow
		Comments []query.Comment
		User     auth.User
	}

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		once.Do(func() {
			tpl = template.Must(template.ParseFS(templateFS, "templates/base.tpl", "templates/details.tpl"))
		})

		if r.Method == "POST" {
			h.detailsPost(w, r)
			return
		}

		if r.Method != "GET" {
			httpError(w, r, "bad method", errors.New(r.Method), http.StatusMethodNotAllowed)
			return
		}

		args := r.URL.Query()

		id, err := strconv.Atoi(args.Get("id"))
		if err != nil {
			httpError(w, r, "can't parse id", err, http.StatusBadRequest)
			return
		}

		data := templateData{
			User: auth.Get(r),
		}

		err = h.Q.RunTx(r.Context(), func(ctx context.Context, q *query.Queries) error {
			data.Wine, err = q.LoadWine(ctx, int32(id))
			if err != nil {
				httpError(w, r, "can't load wine data", err, http.StatusInternalServerError)
				return err
			}

			data.Comments, err = q.ListComments(ctx, int32(id))
			if err != nil {
				httpError(w, r, "can't load comments", err, http.StatusInternalServerError)
				return err
			}

			return nil
		})
		if err != nil {
			return
		}

		err = tpl.ExecuteTemplate(w, "details.tpl", data)
		if err != nil {
			level.Error(log.Get(r)).
				Log("error", err,
					"msg", "can't execute details template")

			return
		}
	})
}
