package storage

import (
	"bytes"
	"context"
	"database/sql"
	"embed"
	"encoding/csv"
	"errors"
	"fmt"
	"image"
	"image/png"
	"io"
	"io/fs"
	"os"
	"path/filepath"
	"sort"
	"strconv"
	"strings"

	"git.c3pb.de/gbe/invinoveritas/log"
	"git.c3pb.de/gbe/invinoveritas/storage/query"
	kitlog "github.com/go-kit/kit/log"
	"github.com/go-kit/kit/log/level"
	"golang.org/x/image/draw"
)

// AddPicture loads picture data (PNG or JPEG) from fh and persists a scaled down version in q.
// If something goes wrong during loading, or the image is neither PNG nor JPEG, an error
// is returned. If contentType is not the empty string, it is validated to be either
// image/png or image/jpeg.
func AddPicture(ctx context.Context, q *query.Queries, wineID int, fh io.Reader, contentType string) (err error) {
	// Encode scaled image as PNG, will contain image data if there is any
	switch contentType {
	case "", "image/jpeg", "image/png":
	default:
		return fmt.Errorf("unexpected content type for image: %q", contentType)
	}

	img, _, err := image.Decode(fh)
	if err != nil {
		return fmt.Errorf("decoding image: %w", err)
	}

	// Scale image down and encode as PNG
	// Get aspect ratio of incoming picture
	bounds := img.Bounds()
	aspect := float64(bounds.Max.X) / float64(bounds.Max.Y)

	const destHeight = 800

	rect := image.Rect(0, 0, int(destHeight*aspect), destHeight)

	level.Info(log.GetContext(ctx)).
		Log("bounds", bounds,
			"aspect", aspect,
			"rect", rect,
			"msg", "resizing image")

	scaled := image.NewRGBA(rect)

	draw.ApproxBiLinear.Scale(scaled, rect, img, bounds, draw.Over, nil)

	var buf bytes.Buffer
	err = png.Encode(&buf, scaled)
	if err != nil {
		return fmt.Errorf("encoding image: %w", err)
	}

	err = q.StorePicture(ctx, query.StorePictureParams{
		WineID:  int32(wineID),
		Picture: buf.Bytes(),
	})
	if err != nil {
		return fmt.Errorf("storing image: %w", err)
	}

	return nil
}

//go:embed migrations/*.sql
var migrationFS embed.FS

func Open(ctx context.Context, dbPath, dumpPath string, logger kitlog.Logger) (*sql.DB, error) {
	db, err := sql.Open("sqlite", dbPath)
	if err != nil {
		return nil, err
	}

	level.Debug(logger).Log("db_path", dbPath)

	pragmas := []string{
		"PRAGMA busy_timeout = 600000",
		"PRAGMA journal_mode = WAL",
	}

	for _, p := range pragmas {
		_, err = db.ExecContext(ctx, p)
		if err != nil {
			return nil, fmt.Errorf("running %q: %w", p, err)
		}
	}

	// Make sure that we have a table for the migrations, the rest runs in a transaction.
	_, err = db.ExecContext(ctx, `CREATE TABLE IF NOT EXISTS migrations (name text, unique(name))`)
	if err != nil {
		return nil, fmt.Errorf("creating migrations table: %w", err)
	}

	q := query.New(db)

	entries, err := migrationFS.ReadDir("migrations")
	if err != nil {
		return nil, fmt.Errorf("reading embedded migration FS: %w", err)
	}

	sort.Slice(entries, func(i, j int) bool {
		return entries[i].Name() < entries[j].Name()
	})

	err = q.RunTx(ctx, func(ctx context.Context, q *query.Queries) error {
		for _, e := range entries {
			// Check if we need to apply that migration
			count, err := q.CountMigrations(ctx, e.Name())
			if err != nil {
				return fmt.Errorf("checking for migration %q: %w", e.Name(), err)
			}

			if count > 1 {
				return fmt.Errorf("migration %q applied more than once", e.Name())
			}

			if count == 1 {
				continue
			}

			content, err := fs.ReadFile(migrationFS, "migrations/"+e.Name())
			if err != nil {
				return fmt.Errorf("reading migration %q: %w", e.Name(), err)
			}

			_, err = q.Raw().ExecContext(ctx, string(content))
			if err != nil {
				return fmt.Errorf("applying migration %q: %w", e.Name(), err)
			}

			// Mark migration as applied
			err = q.AddMigration(ctx, e.Name())
			if err != nil {
				return fmt.Errorf("marking migration %q as applied: %w", e.Name(), err)
			}
		}

		return nil
	})
	if err != nil {
		return nil, fmt.Errorf("applying migrations: %w", err)
	}

	err = q.RunTx(ctx, func(ctx context.Context, q *query.Queries) error {
		v, err := q.GetState(ctx, "migrated-from-ql")
		if err != nil && !errors.Is(err, sql.ErrNoRows) {
			return err
		}
		if v == "true" {
			level.Info(logger).Log("msg", "dump already imported")
			return nil
		}

		level.Info(logger).Log("msg", "attempting import of dumped db", "path", dumpPath)

		winesFH, err := os.Open(filepath.Join(dumpPath, "wines.csv"))
		if err != nil {
			level.Warn(logger).Log("msg", "can't open wines dump, aborting import", "error", err)
			return nil
		}
		defer winesFH.Close()

		commentsFH, err := os.Open(filepath.Join(dumpPath, "comments.csv"))
		if err != nil {
			level.Warn(logger).Log("msg", "can't open comments dump, aborting import", "error", err)
			return nil
		}
		defer commentsFH.Close()

		// Import wines first
		r := csv.NewReader(winesFH)
		r.TrimLeadingSpace = true

		wines := make(map[int]int64) // Maps old ID to new ID

		for {
			record, err := r.Read()
			if errors.Is(err, io.EOF) {
				break
			}
			if err != nil {
				return err
			}
			if len(record) != 5 {
				return fmt.Errorf("unexpected record length %d", len(record))
			}

			rawID, name, rawRating := record[0], record[1], record[2]

			id, err := strconv.Atoi(rawID)
			if err != nil {
				return err
			}

			country := record[3]

			img, err := decodeImg(record[4])
			if err != nil {
				return err
			}

			rating, err := strconv.Atoi(rawRating)
			if rawRating != "" && err != nil {
				return err
			}

			level.Debug(logger).Log("id", id, "name", name, "rating", rating, "country", country, "img_bytes", len(img))
			res, err := q.InsertWine(ctx, query.InsertWineParams{
				Name: name,
				Rating: sql.NullInt32{
					Int32: int32(rating),
					Valid: rawRating != "",
				},
				Country: sql.NullString{
					String: country,
					Valid:  country != "",
				},
			})
			if err != nil {
				return err
			}

			newID, err := res.LastInsertId()
			if err != nil {
				return err
			}

			wines[id] = newID

			err = q.StorePicture(ctx, query.StorePictureParams{
				WineID:  int32(newID),
				Picture: img,
			})
			if err != nil {
				return err
			}
		}

		// And then the comments
		r = csv.NewReader(commentsFH)
		r.TrimLeadingSpace = true

		for {
			record, err := r.Read()
			if errors.Is(err, io.EOF) {
				break
			}
			if err != nil {
				return err
			}
			if len(record) != 3 {
				return fmt.Errorf("unexpected record length %d", len(record))
			}

			rawWineID, comment := record[1], record[2]

			wineID, err := strconv.Atoi(rawWineID)
			if err != nil {
				return err
			}

			level.Debug(logger).Log("wine_id", wines[wineID], "comment", comment)

			err = q.AddComment(ctx, query.AddCommentParams{
				WineID:  int32(wines[wineID]),
				Comment: comment,
			})
			if err != nil {
				return err
			}
		}

		err = q.SetState(ctx, query.SetStateParams{
			Key: "migrated-from-ql",
			Val: "true",
		})
		if err != nil {
			return err
		}

		return nil
	})
	if err != nil {
		return nil, fmt.Errorf("importing dumped DB: %w", err)
	}

	return db, nil
}

func decodeImg(enc string) ([]byte, error) {
	if enc == "<nil>" {
		return nil, nil
	}

	if len(enc) < 3 {
		return nil, errors.New("short encoding")
	}

	if enc[0] != '[' || enc[len(enc)-1] != ']' {
		return nil, errors.New("malformed header/trailer")
	}

	parts := strings.Split(enc[1:len(enc)-1], " ")
	decoded := make([]byte, len(parts))
	for i, p := range parts {
		b, err := strconv.Atoi(p)
		if err != nil {
			return nil, err
		}

		if b < 0 || b > 255 {
			return nil, fmt.Errorf("out of range: %d", b)
		}

		decoded[i] = byte(b)
	}

	return decoded, nil
}
